(function ($, Drupal, drupalSettings) {
    /**
     * Disable the continue buttons in the checkout process once they are clicked
     * and provide a notification to the user.
     */
    Drupal.behaviors.commerceStripeWebPaymentRequest = {
        attach: function (context) {
            var walletType = 'googlePay';
            var shippingAddress = [];
            var initialTotal = drupalSettings.stripe_payment_request.paymentRequest.total.amount;
            var needsWindowReload = false;
            function setWalletType(paymentRequest) {
                if (paymentRequest._buttonTypeName.toLowerCase().includes('apple')) {
                    walletType = 'applePay';
                }
                else if (paymentRequest._buttonTypeName.toLowerCase().includes('link')) {
                    walletType = 'link';
                }
            }

            var $form = $(drupalSettings.stripe_payment_request.form_id);
            // Check if Stripe Payment Request tab active.

            const paymentRequestSettings =
                drupalSettings.stripe_payment_request.paymentRequest;

            // Create our Payment Request button.
            var stripe = Stripe(
                drupalSettings.stripe_payment_request.publishable_key
            );
            var paymentRequest = stripe.paymentRequest(paymentRequestSettings);

            $form.once("stripe-payment-request-setup-wrapper").each(function () {
                var elements = stripe.elements();
                var prButton = elements.create("paymentRequestButton", {
                    paymentRequest: paymentRequest,
                    style: {
                        paymentRequestButton: {
                            height: "44px",
                            theme: "light-outline",
                            type: "buy",
                        },
                    },
                });

                // Check if we can make payments.
                (async () => {
                    const result = await paymentRequest.canMakePayment();
                    // If we can make Stripe Apple/Google payments, show the button.
                    // Only NULL means you can't use it, return with applePay true/false is only meant for apple pay
                    if (result) {
                        prButton.mount("#payment-request-button");
                        setWalletType(paymentRequest);
                        //Drupal.behaviors.commerceStripeWebPaymentRequest.showButton();
                    }
                    // Else, if they can't make payments, show the setup info.
                    else {
                        document.getElementById("payment-request-button").style.display =
                            "none";
                        //Drupal.behaviors.commerceStripeWebPaymentRequest.showSetUpButton();
                    }
                })();

                // Send paymentMethod to server

                paymentRequest.on("paymentmethod", function (ev) {
                    var payment_intent_data = {
                        payment_method_remote_id: ev.paymentMethod.id,
                        result: ev,
                        payment_request:
                            drupalSettings.stripe_payment_request.paymentRequest,
                    };

                    var payment_intent_path =
                        drupalSettings.path.baseUrl +
                        drupalSettings.path.pathPrefix +
                        drupalSettings.stripe_payment_request.payment_intent_path;
                    fetch(payment_intent_path, {
                        method: "POST",
                        body: JSON.stringify({
                            data: payment_intent_data,
                        }),
                        headers: {
                            "content-type": "application/json",
                        },
                    }).then(function (response) {
                        response.json().then(function (json) {
                            if (drupalSettings.stripe_payment_request.recalculate_total_price && json.updated_price) {
                                drupalSettings.stripe_payment_request.paymentRequest.total.amount =
                                    json.updated_price;

                                // Hack to force closing of payment request interface
                                if (walletType === "googlePay") {
                                    ev.complete("success");
                                } else {
                                    paymentRequest.abort();
                                }
                                if (
                                    typeof drupalSettings.stripe_payment_request
                                        .view_summary_selector !== "undefined" &&
                                    drupalSettings.stripe_payment_request
                                        .view_summary_selector === ""
                                ) {
                                    let sidebar = $(
                                        drupalSettings.stripe_payment_request.view_summary_selector
                                    ).trigger("RefreshView")[0];
                                    $(
                                        drupalSettings.stripe_payment_request.view_summary_selector
                                    ).html(sidebar);
                                }

                                let errorMessage = Drupal.t(drupalSettings.stripe_payment_request.recalculate_total_price_error_message);
                                setTimeout(() => {
                                    alert(errorMessage);
                                    window.location.reload();
                                }, 500);
                            } else if (json.error) {
                                Drupal.commerceStripe.displayError(result.error.message);
                                return false;
                            } else if (json.requires_action) {
                                ev.complete("success");
                                handleAction(json, payment_intent_path, ev);
                            } else if (!json.success) {
                                ev.complete("success");
                                window.location.href = json.redirect_uri;
                            }
                            else {
                                ev.complete("success");
                                window.location.href = json.redirect_uri;
                            }
                        });
                    })
                        .catch(error => {
                            console.error('There was a problem with the fetch operation:', error.message);
                            // You can also display an error message or take other appropriate actions here.
                        });
                });

                function handleAction(response, payment_intent_path, ev) {
                    var data = {
                        payment_method: response.payment_method_remote_id,
                    };
                    stripe
                        .handleCardPayment(response.payment_intent_client_secret, data)
                        .then(function (result) {
                            if (result.error) {
                                Drupal.commerceStripe.displayError(result.error.message);
                            } else {
                                //We need to send back info to controller and to add new data to order and payment
                                fetch(payment_intent_path, {
                                    method: "POST",
                                    body: JSON.stringify({
                                        data: result,
                                    }),
                                    headers: {
                                        "content-type": "application/json",
                                    },
                                }).then(function (response) {
                                    response.json().then(function (json) {
                                        if (json.error) {
                                            Drupal.commerceStripe.displayError(result.error.message);
                                            return false;
                                        } else if (json.requires_action) {
                                            ev.complete("success");
                                            handleAction(json, payment_intent_path, ev);
                                        } else {
                                            ev.complete("success");
                                            window.location.href = json.redirect_uri;
                                        }
                                    });
                                });
                            }
                        })
                        .catch(error => {
                            console.error('There was a problem with the fetch operation:', error.message);
                            // You can also display an error message or take other appropriate actions here.
                        });
                }

                paymentRequest.on("shippingaddresschange", function (ev) {
                    // Perform server-side request to fetch shipping options

                    var shipping_path =
                        drupalSettings.path.baseUrl +
                        drupalSettings.path.pathPrefix +
                        drupalSettings.stripe_payment_request.shipping_path;

                    shippingAddress = ev.shippingAddress;

                    fetch(shipping_path, {
                        method: "POST",
                        body: JSON.stringify({
                            shippingAddress: ev.shippingAddress,
                        }),
                        headers: {
                            "content-type": "application/json",
                        },
                    })
                        .then(function (response) {
                            return response.json();
                        })
                        .then(function (result) {
                            if (result.shippingOptions && result.shippingOptions.length) {
                                ev.updateWith({
                                    status: "success",
                                    shippingOptions: result.shippingOptions,
                                });
                            } else {
                                ev.updateWith({
                                    status: "invalid_shipping_address",
                                });
                            }
                        })
                        .catch(error => {
                            console.error('There was a problem with the fetch operation:', error.message);
                            // You can also display an error message or take other appropriate actions here.
                        });
                });

                function shippingOptionChange(event) {
                    var updateWith = event.updateWith;
                    // handle shippingoptionchange event
                    var updateShipmentPath = '/stripe-payment-request-button/update-shipping/' + drupalSettings.stripe_payment_request.order_uuid;
                    updateShipmentPath += '/' + event.shippingOption.id;
                    fetch(updateShipmentPath, {
                        method: "POST",
                        body: JSON.stringify({
                            shippingAddress: shippingAddress,
                        }),
                        headers: {
                            "content-type": "application/json",
                        },
                    })
                        .then(function (response) {
                            return response.json();
                        })
                        .then(function (result) {
                            if (result.success === false) {
                                Drupal.commerceStripeButton.displayError(result.error);
                                return false;
                            }
                            else {
                                if (initialTotal !== result.amount) {
                                    needsWindowReload = true;
                                }
                                // Update paymentRequest with the new shipping option
                                updateWith({
                                    status: "success",
                                    total: {
                                        amount: result.amount,
                                        'label': 'Total due',
                                    },
                                    shippingOptions: result.shippingOptions
                                });
                            }
                        })
                        .catch(error => {
                            console.error('There was a problem with the fetch operation:', error.message);
                            // You can also display an error message or take other appropriate actions here.
                        });
                }
                paymentRequest.on('shippingoptionchange', function (event) {
                    shippingOptionChange(event);
                });

                paymentRequest.on('cancel', function () {
                    if (needsWindowReload === true) {
                        setTimeout(() => {
                            window.location.reload();
                        }, 500);
                    }
                });

            });
        },
    };

    Drupal.commerceStripeButton = {
        displayError: function (errorMessage) {
            $("#payment-errors").html(
                Drupal.theme("commerceStripeButtonError", errorMessage)
            );
        },
    };
    Drupal.theme.commerceStripeButtonError = function (message) {
        return $('<div class="messages messages--error"></div>').html(message);
    };
})(jQuery, Drupal, drupalSettings);
