<?php

namespace Drupal\commerce_stripe_payment_request_button\Controller;

use Stripe\Stripe;
use Stripe\PaymentIntent;
use Stripe\Exception\ApiErrorException;

use Drupal\Core\Controller\ControllerBase;
use Drupal\commerce_stripe_payment_request_button\Services\PaymentRequestButtonHelperServiceInterface;
use Drupal\commerce_stripe\ErrorHelper;
use Drupal\Core\Entity\EntityInterface;
use Drupal\commerce_payment\Entity\Payment;
use Drupal\Component\Serialization\Json;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Url;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Drupal\Core\Access\AccessResult;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal;
use Drupal\commerce_order\Entity\Order;
use Drupal\commerce_stripe_payment_request_button\Plugin\Commerce\PaymentGateway\PaymentRequestButtonInterface;
use Exception;
use Drupal\commerce_order\Event\OrderEvent;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;
use Drupal\commerce_stripe_payment_request_button\Event\OrderEvents;
use Drupal\commerce_shipping\Entity\ShipmentInterface;
use Drupal\commerce_shipping\Entity\Shipment;
use Drupal\commerce_shipping\Entity\ShippingMethod;


class PaymentRequestButtonCheckout extends ControllerBase {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The Payment Request button service.
   *
   * @var \Drupal\commerce_stripe_payment_request_button\Services\PaymentRequestButtonHelperServiceInterface
   */
  protected $paymentRequestButtonService;

  /**
   * The messenger.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * The messenger.
   *
   * @var \Drupal\Core\Routing\RouteMatchInterface
   */
  protected $route_match;

  /**
   * Creates a new ViewModeAccessCheck.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   */
  public function __construct(
    EntityTypeManagerInterface                 $entity_type_manager,
    PaymentRequestButtonHelperServiceInterface $paymentRequestButtonService,
    MessengerInterface                         $messenger,
    RouteMatchInterface                        $route_match
  ) {
    $this->entityTypeManager = $entity_type_manager;
    $this->paymentRequestButtonService = $paymentRequestButtonService;
    $this->messenger = $messenger;
    $this->route_match = $route_match;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('commerce_stripe_payment_request_button.helper'),
      $container->get('messenger'),
      $container->get('current_route_match')
    );
  }

  /**
   * Handles the incoming order from the "classic" checkout process
   *
   * @param EntityInterface $commerce_order
   * @param Request $request
   *
   * @return JsonResponse
   */
  public function handleOrder(EntityInterface $commerce_order, Request $request) {
    // Transform JSON values to array
    $values = Json::decode($request->getContent());
    $shipping_info = '';

    if (isset($values["data"]["result"]["shippingOption"]["shippingAddress"])) { //if multiple shipping addresses are available
      $shipping_info = $values["data"]["result"]["shippingOption"]["shippingAddress"];
    } elseif (isset($values["data"]["result"]["shippingAddress"])) {
      $shipping_info = $values["data"]["result"]["shippingAddress"];
    }
    $order = Order::load($commerce_order->id());
    $intialTotalPrice = $order->getTotalPrice()->getNumber();
    //get the current currency order
    $currency = strtolower($order->getTotalPrice()->getCurrencyCode());

    // Get the configuration of the payment gateway
    $orderPaymentConfiguration =
      $this->paymentRequestButtonService->getOrderPaymentConfiguration($order, $currency);

    $configName = $orderPaymentConfiguration->configName;
    $config = $orderPaymentConfiguration->config;
    Stripe::setApiKey($config['secret_key']);
    $billingProfile = $this->paymentRequestButtonService->assignOrCreateProfile($order, $values, 'billing');
    $this->paymentRequestButtonService->assignProfileToOrder($order, $billingProfile, 'billing');
    // Deal with shipping stuff only if we selected that we want to fetch shipping address
    if (
      $config['request_shipping'] && !empty($shipping_info)
    ) {
      $shippingProfile = $this->paymentRequestButtonService->assignOrCreateProfile($order, $values, 'shipping');
      $this->paymentRequestButtonService->assignProfileToOrder(
        $order,
        $shippingProfile,
        'shipping',
        $values
      );
    }

    $relcalculate = $config['recalculate_total_price'];
    if ($relcalculate) {
      /** @var \Symfony\Component\EventDispatcher\EventDispatcherInterface $eventDispatcher */
      $eventDispatcher = \Drupal::service('event_dispatcher');

      // Dispatch the ORDER_UPDATE event.
      $orderEvent = new OrderEvent($order);
      $eventDispatcher->dispatch($orderEvent, OrderEvents::ORDER_RECALCULATE);
      //in the case of recalculation we need to reload the order
      $order = Order::load($order->id());
    }
    //intentinally left variable for easier debugging
    $recalculatedNumber = $order->getTotalPrice()->getNumber();
    $priceCheck = ($relcalculate && $intialTotalPrice != $recalculatedNumber);

    // If order has problems, take it back to cart
    if (empty($order) || $order->getState()->value == 'canceled' || $priceCheck) {
      $response = [
        'success' => false,
        'redirect_uri' => Url::fromRoute(Url::fromRoute('commerce_cart.page')
          ->toString()),
      ];
      if ($priceCheck) {
        $response = [
          'success' => true,
          'updated_price' => $this->paymentRequestButtonService->toMinorUnits($order->getTotalPrice()),
        ];
      }

      return new JsonResponse($response);
    }

    // Check if we already have payment intent
    $intentId = $order->getData('stripe_intent');
    if (!is_null($intentId)) {
      try {
        $paymentIntent = $this->paymentRequestButtonService->updateDrupalPaymentByPaymentIntent($order, $intentId);
      } catch (ApiErrorException $e) {
        ErrorHelper::handleException($e);
      }
    } else {
      try {
        $shippingPhone = $values["data"]['result']['payerPhone'];
        $shippingName = $values["data"]['result']['payerName'];
        if (
          isset($values["data"]['result']['shippingAddress']['recipient']) &&
          !empty($values["data"]['result']['shippingAddress']['recipient'])
        ) {
          $shippingName = $values["data"]['result']['shippingAddress']['recipient'];
        }
        if (
          isset($values["data"]['result']['shippingAddress']['phone']) &&
          !empty($values["data"]['result']['shippingAddress']['phone'])
        ) {
          $shippingPhone = $values["data"]['result']['shippingAddress']['phone'];
        }
        $paymentIntentData = [
          'payment_method' => $values["data"]["payment_method_remote_id"],
          'amount' => $values["data"]["payment_request"]["total"]["amount"],
          'currency' => $values["data"]["payment_request"]["currency"],
          'payment_method_types' => ['card'],
          'capture_method' => $config['capture_method'],
          'description' => 'Order ' . $order->id(),
          'metadata' => ['order_id' => $order->id()],
          'confirmation_method' => 'automatic',
          'confirm' => TRUE,
          'shipping' => [
            'name' => $shippingName,
            'phone' => $shippingPhone,
            'address' => [
              'line1' => $shipping_info['addressLine'][0],
              'line2' => isset($shipping_info['addressLine'][1]) ? $shipping_info['addressLine'][1] : '',
              'city' => $shipping_info['city'],
              'state' => isset($shipping_info['region']) ? $shipping_info['region'] : '',
              'postal_code' => $shipping_info['postalCode'],
              'country' => $shipping_info['country'],
            ],
          ],

        ];
        // payment_request_button_shipping
        // Create payment intent
        $paymentIntent = PaymentIntent::create($paymentIntentData);

        if ($order->hasItems()) {
          // Put order in proper status and assign value
          $shipping = '';
          if (isset($values['data']['result']['shippingOption']['id'])) {
            $shipping = $values['data']['result']['shippingOption']['id'];
          }
          $this->paymentRequestButtonService->setOrderData($order, $paymentIntent, $values["data"]["result"]["payerEmail"], $configName, $shipping);

          // Create payment for this order and give it values
          $payment_gateway = $this->entityTypeManager
            ->getStorage('commerce_payment_gateway')
            ->load($configName);

          $current_user = $order->getCustomerId();

          $payment = $this->paymentRequestButtonService->createDrupalPayment($order, $payment_gateway, $paymentIntent, $current_user, $config['capture_method']);

          $this->paymentRequestButtonService->processProfile($order, $paymentIntent, $config['request_phone']);
        }
      } catch (ApiErrorException $e) {
        // Server will return 500 to the frontend.
        $this->messenger->addError($e->getMessage());
        ErrorHelper::handleException($e);
      }
    }

    if ($paymentIntent->status == 'requires_confirmation' || $paymentIntent->status == 'requires_source_action' || $paymentIntent->status == 'requires_action') {
      # Tell the client to handle the action
      $response = [
        'success' => true,
        'requires_action' => TRUE,
        'payment_intent_client_secret' => $paymentIntent->client_secret,
        'payment_method_remote_id' => $values["data"]["payment_method_remote_id"],
        'redirect_uri' => Url::fromRoute('commerce_checkout.form', ['commerce_order' => $order->id()])
            ->toString() . '/payment',
      ];
    } else {
      if ($paymentIntent->status == 'succeeded' || $paymentIntent->status == 'requires_capture') {
        # The payment didn’t need any additional actions and completed!
        # Handle post-payment fulfillment
        $response = [
          'success' => TRUE,
          'redirect_uri' => Url::fromRoute('commerce_checkout.form', ['commerce_order' => $order->id(),])
              ->toString() . '/payment',
        ];
      } else {
        # Invalid status
        //http_response_code(500);
        $response = [
          'success' => false,
          'error' => 'Invalid PaymentIntent status'
        ];
      }
    }
    return new JsonResponse($response);
  }

  //check if order marked as paid
  //check if payment intent paid

  public function handleDecoupledOrder($commerceOrderUuId, $payment_intent) {
    $orders = $this->entityTypeManager->getStorage('commerce_order')
      ->loadByProperties(['uuid' => $commerceOrderUuId]);
    $order = reset($orders);

    //get the current currency order

    $currency = strtolower($order->getTotalPrice()->getCurrencyCode());

    $orderPaymentConfiguration =
      $this->paymentRequestButtonService->getOrderPaymentConfiguration($order, $currency);

    // Init Stripe and config
    $configName = $orderPaymentConfiguration->configName;
    $config = $orderPaymentConfiguration->config;

    Stripe::setApiKey($config['secret_key']);

    $paymentIntent = PaymentIntent::retrieve($payment_intent);

    if ($order->id() == $paymentIntent->metadata->order_id) {
      // If order has problems, take it back to cart
      if (empty($order) || $order->getState()->value == 'canceled') {
        $response = [
          'redirect_uri' => Url::fromRoute(Url::fromRoute('commerce_cart.page')
            ->toString()),
        ];
        return new JsonResponse($response);
      }

      $payments = \Drupal::entityQuery('commerce_payment')
        ->condition('order_id', $order->id())
        ->execute();

      $paymentCharge = $paymentIntent->charges['data'][0];
      $shipping_info = '';
      if (isset($paymentCharge->shipping->address)) { //if multiple shipping addresses are available
        $shipping_info = $paymentCharge->shipping->address;
      }
      if (count($payments)) {
        $payment = Payment::load(reset($payments)); //cannot flip
        $payment->setRemoteId($paymentIntent->id);
        $payment->save();
      } else {
        if ($order->hasItems()) {
          // Put order in proper status and assign values
          $this->paymentRequestButtonService->setOrderData($order, $paymentIntent, $paymentIntent->metadata->email, $configName);

          // Create payment for this order and give it values
          $payment_gateway = $this->entityTypeManager
            ->getStorage('commerce_payment_gateway')
            ->load($configName);

          $current_user = $order->getCustomerId();
          $payment = $this->paymentRequestButtonService->createDrupalPayment($order, $payment_gateway, $paymentIntent, $current_user, $config['capture_method']);
        }
      }
      // Deal with billing profile
      $this->paymentRequestButtonService->processProfile($order, $paymentIntent, $config['request_phone']);

      // Deal with shipping stuff only if we selected that we want to fetch shipping address
      if ($config['request_shipping'] && !empty($shipping_info)) {
        $this->paymentRequestButtonService->handleShipping($order, $paymentIntent, $config['request_phone']);
      }

      if ($paymentIntent->status == PaymentIntent::STATUS_SUCCEEDED) {
        if (!$payment->isCompleted()) {
          $workflow = $payment->getState()->getWorkflow();
          $transition = $workflow->getTransition('capture');
          $payment->getState()->applyTransition($transition);
          $payment->save();
        }
        $response = [
          'success' => TRUE,
          'redirect_uri' => Url::fromRoute('commerce_checkout.form', ['commerce_order' => $order->id()])
              ->toString() . '/payment',
        ];
      }
    } else {
      //orders do not match
      $response = [
        'success' => FALSE,
        'error' => $this->t('Order id from the PaymentIntent does not match our records'),
      ];
    }
    return new JsonResponse($response);
  }

  public function updateOrderShipping($commerceOrderUuId, $shippingId, Request $request) {
    $orders = $this->entityTypeManager->getStorage('commerce_order')
      ->loadByProperties(['uuid' => $commerceOrderUuId]);
    $order = reset($orders);
    $values = Json::decode($request->getContent());
    try {
      $shipment = $this->paymentRequestButtonService->getShipmentByShippingMethod($order, $shippingId);
      $data['data']['result']['shippingOption']['id'] = $shippingId;
      //has shipment and shipping method id different from the one we are trying to set
      if (is_null($shipment)) {
        $shipment = $this->paymentRequestButtonService->createShipment($order, $data);
      }

      $this->paymentRequestButtonService->setShipment($shipment, $shippingId);
      $order->set('shipments', $shipment);
      $order->save();
      $orderTotal = $order->getTotalPrice();
      $shippingOptions = $this->paymentRequestButtonService->getShippingOptions($order, $values['shippingAddress']);

      return new JsonResponse([
        'success' => TRUE, 'amount' => $this->paymentRequestButtonService->toMinorUnits($orderTotal), 'currency' => $orderTotal->getCurrencyCode(),
        'shippingOptions' => $shippingOptions
      ]);
    } catch (Exception $e) {
      $error_message = $e->getMessage();
      $this->messenger->addError($error_message, TRUE);
      Drupal::logger('commerce_stripe_payment_request_button')
        ->error(t('Failed to update order @order_id shippiping . @message', [
          '@order_id' => $order->id(),
          '@message' => $e,
        ]));
    }
    return new JsonResponse(['success' => FALSE]);
  }

  /**
   * Checks access for the form page.
   *
   * @return \Drupal\Core\Access\AccessResult
   *   The access result.
   */
  public function checkAccess() {
    $orderCollection = $this->entityTypeManager->getStorage('commerce_order')
      ->loadByProperties(['uuid' => $this->route_match->getParameter('commerceOrderUuId')]);
    $order = reset($orderCollection);
    if (!$order || ($order && $order->getState()->getId() == 'canceled')) {
      return AccessResult::forbidden()->addCacheableDependency($order);
    }

    $access = AccessResult::allowedIf($order->hasItems())
      ->addCacheableDependency($order);
    return $access;
  }
}
