<?php

namespace Drupal\commerce_stripe_payment_request_button\Services;

use Drupal;
use Exception;
use Stripe\Stripe;
use Drupal\commerce_stripe_payment_request_button\Services\PaymentRequestButtonHelperServiceInterface;
use Stripe\PaymentIntent;
use Stripe\PaymentMethod;
use Stripe\Exception\ApiErrorException;
use Drupal\Core\Locale\CountryManager;
use Drupal\profile\Entity\ProfileInterface;
use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_payment\Entity\Payment;
use Drupal\commerce_price\Calculator;
use Drupal\commerce_price\Price;
use Drupal\commerce_shipping\Entity\Shipment;
use Drupal\commerce_shipping\ShipmentItem;
use Drupal\commerce_store\Entity\Store;

use Drupal\physical\Weight;
use Drupal\physical\WeightUnit;
use Drupal\profile\Entity\Profile;
use Drupal\commerce_payment\Entity\PaymentGatewayInterface;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;
use Drupal\commerce_shipping\Event\ShippingEvents;
use Drupal\commerce_shipping\Event\ShippingRatesEvent;

/**
 * Class PaymentRequestButtonHelperService
 *
 * @package Drupal\commerce_stripe_payment_request_button\Service
 */
class PaymentRequestButtonHelperService implements PaymentRequestButtonHelperServiceInterface {

  /**
   * Check if the payment is Google Pay, Apple pay or Link
   *
   * @param PaymentIntent $paymentIntent
   * @return string $wallet
   */
  public function getWalletType(PaymentIntent $paymentIntent): string {
    $wallet = 'Link';

    if (isset($paymentIntent->charges->data[0])) {
      /**
       * @var \Stripe\PaymentIntent $paymentIntent
       */
      $charges = $paymentIntent->charges->data[0];
      $charges_values = $charges->jsonSerialize();
      $walletTypeSet = isset($charges_values['payment_method_details']['card']['wallet']['type']);
      if ($walletTypeSet == TRUE) {
        $walletType = $charges_values['payment_method_details']['card']['wallet']['type'];
        if (strpos($walletType, 'apple') !== false) {
          $wallet = 'Apple pay';
        }
        if (strpos($walletType, 'google') !== false) {
          $wallet = 'Google pay';
        }
      }
    }
    return $wallet;
  }

  /**
   * Get the order configuration by currency
   *
   * @param string $currency
   * @return class $data
   */
  public function getOrderPaymentConfiguration(OrderInterface $order, string $currency) {
    $configCurrency = $order->getData('stripe_payment_request_button_config_id');
    $configBase = 'commerce_stripe_payment_request_button';

    if (!$configCurrency) {
      $configCurrency = $configBase . '_' . $currency;
    }
    $configName = $configCurrency;

    // Load curency config
    $config = \Drupal::config('commerce_payment.commerce_payment_gateway.' . $configCurrency);

    if (!$config->get('status')) {
      // Load default config
      $config = \Drupal::config('commerce_payment.commerce_payment_gateway.' . $configBase);
      $configName = $configBase;
    }

    $data = new \stdClass();

    $data->config = $config->get('configuration');
    $data->configName = $configName;

    return $data;
  }

  /**
   * Update drupal payment from the payment intent
   *
   * @param OrderInterface $order
   * @param string $intentId
   * @return PaymentIntent $paymentIntent
   */
  public function updateDrupalPaymentByPaymentIntent(OrderInterface $order, string $intentId) {

    $paymentIntent = PaymentIntent::retrieve($intentId);
    $payments = \Drupal::entityQuery('commerce_payment')
      ->condition('order_id', $order->id())
      ->execute();

    $payment = Payment::load(reset($payments));
    $payment->setRemoteId($paymentIntent->id);
    $payment->save();

    // Reset this step as it gets to NULL on cart refresh
    $order->get('checkout_flow')->setValue("payment_request_button_flow");
    $order->get('checkout_step')->setValue("payment");
    $order->save();
    return  $paymentIntent;
  }

  /**
   * Set the following order data: checkout_flow,checkout_step,email,payment_gateway,stripe_intent,wallet
   *
   * @param OrderInterface $order
   * @param PaymentIntent $paymentIntent
   * @param string $email
   * @param string $configName
   * @param string $shipping
   * @return void
   */
  public function setOrderData(OrderInterface $order, PaymentIntent $paymentIntent, string $email, string $configName, string $shipping = ''): void {
    $order->get('checkout_flow')->setValue("payment_request_button_flow");
    $order->get('checkout_step')->setValue("payment");
    $order->setEmail($email);
    $order->get('payment_gateway')->setValue($configName);
    $order->setData('stripe_intent', $paymentIntent->id);
    $order->setData('wallet', $this->getWalletType($paymentIntent));
    if (!empty($shipping)) {
      $order->setData('payment_request_button_shipping', $shipping);
    }
    $order->save();
  }


  /**
   * Create a drupal payment from the data provided
   *
   * @param OrderInterface $order
   * @param PaymentGatewayInterface $paymentGateway
   * @param PaymentIntent $paymentIntent
   * @param string $currentUser
   * @return Payment $payment
   */
  public function createDrupalPayment(OrderInterface $order, PaymentGatewayInterface $paymentGateway, PaymentIntent $paymentIntent, string $currentUser) {
    $paymentMethod = $this->createPaymentMethod($order, $paymentGateway, $paymentIntent->payment_method);
    $this->attachCustomer($order, $paymentIntent->payment_method);

    $payment = Payment::create([
      'state' => 'new',
      'amount' => $order->getTotalPrice(),
      'payment_gateway' => $paymentGateway->id(),
      'order_id' => $order->id(),
      'remote_id' => $paymentIntent->id,
      'payment_gateway_mode' => $paymentGateway->getPlugin()->getMode(),
      'payment_method' => $order->get('payment_method'),
      'expires' => 0,
      'uid' => $currentUser,
    ]);
    $payment->save();


    return $payment;
  }


  /**
   * Check if the shipping profile exists or create a new one.
   * Afterwards return the profile for processing
   *
   * @param $order
   * @param $paymentIntent
   * @param $requestPhone
   * @param $payerPhone
   *
   * @return mixed|null
   */
  public function getOrderShippingProfile($order, $paymentIntent, $requestPhone, $payerPhone) {
    $shippingInfo = [];
    $profiles = $order->collectProfiles();
    $billingDetails = $paymentIntent->charges->data[0]->billing_details;
    $shippingDetails = $paymentIntent->shipping;

    // Determine shipping information
    if (!is_null($paymentIntent->shipping) && count($paymentIntent->shipping->address)) {
      $shippingInfo = $paymentIntent->shipping->address;
    } elseif (count($paymentIntent->charges['data'][0])) {
      $shippingInfo = $billingDetails->address;
    }

    // Convert shipping information to object if it's an array
    if (is_array($shippingInfo)) {
      $shippingInfo = json_decode(json_encode($shippingInfo));
    }

    // Handle addressLine to line1 and line2 conversion
    if (!isset($shippingInfo->line1) && isset($shippingInfo->addressLine[0])) {
      $shippingInfo->line1 = $shippingInfo->addressLine[0];
    }
    if (!isset($shippingInfo->line2) && isset($shippingInfo->addressLine[1])) {
      $shippingInfo->line2 = $shippingInfo->addressLine[1];
    }

    $names = [];
    $copyNames = false;

    //if there are no shipping details that means that We have to use the billing name
    if (is_null($shippingDetails)) {
      $copyNames = true;
    } elseif (isset($shippingDetails->name) && strlen(trim($shippingDetails->name) >= 1)) {
      $names = $this->processCustomerName($shippingDetails->name);
      if (empty($names[0]) || empty($names[1])) {
        $copyNames = true;
      }
    }

    // If names need to be copied, use billing details
    if ($copyNames == true && strlen(trim($billingDetails->name)) >= 1) {
      $names = $this->processCustomerName($billingDetails->name);
    } elseif ($copyNames == true) {
      $billingProfileAddress = $order->getBillingProfile()->address->first();
      $names = [
        !is_null($billingProfileAddress->getGivenName()) ? $billingProfileAddress->getGivenName() : '',
        !is_null($billingProfileAddress->getFamilyName()) ? $billingProfileAddress->getFamilyName() : '',
      ];
    }
    if (isset($shippingInfo->postalCode)) {
      $shippingInfo->postal_code = $shippingInfo->postalCode;
    }
    $given_name = $names['given_name'];
    $family_name = $names['family_name'];

    $shippingProfile = NULL;

    //if the user is authenticated
    if ($order->getCustomerId() != 0) {
      //load all profiles to see if shipping address already exists to prevent database bloat
      $searchData = [
        'type' => 'customer_shipping',
        'uid' => $order->getCustomerId(),
        'address.country_code' => $shippingInfo->country,
        'address.locality' => $shippingInfo->city,
        'address.postal_code' => $shippingInfo->postal_code,
        'address.address_line1' => $shippingInfo->address_line1,
        'address.given_name' => $given_name,
        'address.family_name' => $family_name,

      ];
      if (isset($shippingInfo->region)) {
        $searchData['address.administrative_area'] = $shippingInfo->region;
      }
      if (isset($profiles['shipping'])) {
        $profile = $profiles['shipping'];
      } else {
        $profiles = \Drupal::entityTypeManager()
          ->getStorage('profile')->loadByProperties($searchData);
        $profile = reset($profiles);
      }
      //check if the address exists
      if (!is_bool($profile)) {
        if ($profile->hasField('field_phone') && $requestPhone) {
          $profile->set('field_phone', $payerPhone);
        }
        $profile->save();
        $shippingProfile = $profile;
      }
    }
    $orderShipments = $order->shipments->referencedEntities();

    $firstShipment = reset($orderShipments);

    //if profile exists set it as the order shipping profile else create new order
    if (!is_null($shippingProfile)) {
      if ($shippingProfile->hasField('field_phone') && $requestPhone) {
        $shippingProfile->set('field_phone', $payerPhone);
      }
      $shippingProfile->save();
      $firstShipment->setShippingProfile($shippingProfile);
    } else {
      $createData =  [
        'type' => 'customer_shipping',
      ];
      if ($order->getCustomerId() > 0) {
        $createData['uid'] = $order->getCustomerId();
      }

      $shippingProfile = Profile::create($createData);
      $shippingProfile->address->given_name = $given_name;
      $shippingProfile->address->family_name = $family_name;
      $shippingProfile->address->country_code = $shippingInfo->country;
      $shippingProfile->address->locality = $shippingInfo->city;
      $shippingProfile->address->administrative_area = $shippingInfo->state;
      $shippingProfile->address->postal_code = $shippingInfo->postal_code;
      $shippingProfile->address->address_line1 = $shippingInfo->line1;
      if (isset($shippingInfo->line2)) {
        $shippingProfile->address->address_line2 = $shippingInfo->line2;
      }
      if (isset($shippingInfo->region)) {
        $shippingProfile->address->address_administrative_area = $shippingInfo->region;
      }

      if ($shippingProfile->hasField('field_phone') && $requestPhone) {
        $shippingProfile->set('field_phone', $payerPhone);
      }
      $shippingProfile->save();
    }

    return $shippingProfile;
  }

  public function handleShipping($order, $paymentIntent, $requestPhone = FALSE) {
    $shipment = null;
    $charge = $paymentIntent->charges->data[0];
    if ($order->hasField('shipments') && !$order->get('shipments')->isEmpty()) {
      $shipment = $order->get('shipments')->entity;
    } else {
      $shipment = Shipment::create([
        'type' => 'default',
        'order_id' => $order->id(),
        'title' => 'Shipment',
        'state' => 'ready',
      ]);
    }

    // Loop through order items and add them to shipment
    foreach ($order->getItems() as $order_item) {
      $quantity = $order_item->getQuantity();
      $purchased_entity = $order_item->getPurchasedEntity();

      //TODO Skip weight if there is none, currently breaks order flow
      if ($purchased_entity->get('weight')->isEmpty()) {
        $weight = new Weight(1, WeightUnit::GRAM);
      } else {
        $weight_item = $purchased_entity->get('weight')->first();
        $weight = $weight_item->toMeasurement();
      }

      $shipment_item = new ShipmentItem([
        'order_item_id' => $order_item->id(),
        'title' => $purchased_entity->label(),
        'quantity' => $quantity,
        'weight' => $weight->multiply($quantity),
        'declared_value' => $order_item->getTotalPrice(),
      ]);
      $shipment->addItem($shipment_item);
    }

    $shipping_method = $order->get('shipments')->entity->getShippingMethod();

    if (!is_null($order->getData('payment_request_button_shipping'))) {
      $shipping_method_storage = \Drupal::entityTypeManager()->getStorage('commerce_shipping_method');
      $shipping_method = $shipping_method_storage->load($order->getData('payment_request_button_shipping'));
    } elseif (isset($values['result']['shippingOption']) &&  $shipping_method->id() != $values['data']['result']['shippingOption']['id']) {
      $shipping_method_storage = \Drupal::entityTypeManager()->getStorage('commerce_shipping_method');
      $shipping_method = $shipping_method_storage->load($values['data']['result']['shippingOption']['id']);
    }
    $shipment->setTitle($shipping_method->getName());
    $shipment->setShippingMethodId($shipping_method->id());
    $shipment->setShippingMethod($shipping_method);
    $shipment->save();
    $order->set('shipments', $shipment);

    $order->save();

    // Add proper shipping method
    $shipping_method_plugin = $shipping_method->getPlugin();
    $shipping_rates = $shipping_method_plugin->calculateRates($shipment);
    $shipping_rate = reset($shipping_rates);
    $shipping_service = $shipping_rate->getService()->getId();
    $shipping_method_amount = $shipping_rate->getAmount();

    //if the order currency is different from the default currency, check the if there is an correct shipping field set
    if (
      $shipping_method_amount->getCurrencyCode() != $order->getTotalPrice()
        ->getCurrencyCode()
      && isset($shipping_method_plugin->getConfiguration()['fields'][$order->getTotalPrice()
          ->getCurrencyCode()])
    ) {
      $newAmount = $shipping_method_plugin->getConfiguration()['fields'][$order->getTotalPrice()
        ->getCurrencyCode()];
      if (is_numeric($newAmount['number']) && $newAmount['number'] > 0) {
        $shipping_method_amount = new Price($newAmount['number'], $newAmount['currency_code']);
      }
    }
    //amount is not zero and there are not currency fields appended We use the commerce_currency_resolver to convert the amount
    elseif (
      $shipping_method_amount->getNumber() > 0 &&
      $shipping_method_amount->getCurrencyCode() != $order->getTotalPrice()
        ->getCurrencyCode()
      && isset($shipping_method_plugin->getConfiguration()['fields'][$order->getTotalPrice()
          ->getCurrencyCode()])
    ) {
      $shipping_method_amount = Drupal::service('commerce_currency_resolver.calculator')
        ->priceConversion($shipping_method_amount, $order->getTotalPrice()
          ->getCurrencyCode());
    }

    $shipment->setAmount($shipping_method_amount);

    $phone = '';
    if (is_array($charge) && isset($charge['data']['result']['shippingAddress']['phone'])) {
      $phone = $charge['data']['result']['shippingAddress']['phone'];
    } elseif (!is_array($charge)) { //just to check that it is an object
      $phone = $charge->billing_details->phone;
    }
    $profile = $this->getOrderShippingProfile($order, $paymentIntent, $requestPhone, $phone);

    $shipment->setShippingProfile($profile);
    $shipment->setShippingService($shipping_service);
    $shipment->save();

    $order->set('shipments', $shipment);
    $order->save();
  }

  public function processProfile(OrderInterface $order, $paymentIntent, $requestPhone = true) {
    // Deal with billing profile
    $profile = $order->getBillingProfile();
    $currentUser = $order->getCustomerId();
    $stripeBilling = $paymentIntent->charges['data'][0]->billing_details;
    $stripeShipping = $paymentIntent->shipping;
    if (empty($profile)) {
      $profile = Profile::create([
        'type' => 'customer',
        'uid' => $currentUser,
      ]);
      $profile->save();
    }

    $names = [];
    $billingName = $shippingName = '';

    $billingName = $this->processCustomerName($stripeBilling->name);
    $shippingName = $this->processCustomerName($stripeShipping->name);
    if (empty($billingName[0]) || empty($billingName[1])) {
      $given_name = $shippingName['given_name'];
      $family_name = $shippingName['family_name'];
    } else {
      $given_name = $billingName['given_name'];
      $family_name = $billingName['family_name'];
    }

    $profile->address->given_name = $given_name;
    $profile->address->family_name = $family_name;
    $profile->address->country_code = $stripeBilling->address->country;
    $profile->address->locality = $stripeBilling->address->city;
    $profile->address->adinistrative_area = $stripeBilling->address->state;
    $profile->address->postal_code = $stripeBilling->address->postal_code;
    $profile->address->address_line1 = $stripeBilling->address->line1;
    if (isset($stripeBilling->address->line2)) {
      $profile->address->address_line2 = $stripeBilling->address->line2;
    }
    if ($profile->hasField('field_phone') && $requestPhone) {
      $profile->set('field_phone', $stripeBilling->phone);
    }
    $profile->save();

    $order->setBillingProfile($profile);
    $order->save();
  }

  /**
   * @{inheritdoc}
   */
  public function createPaymentMethod(OrderInterface $order, PaymentGatewayInterface $payment_gateway, string $paymentMethod) {
    $billing_profile = $order->getBillingProfile();
    if ($billing_profile instanceof ProfileInterface) {
      $type = 'stripe_payment_request_button';

      /** @var \Drupal\commerce_payment\PaymentMethodStorageInterface $payment_method_storage */
      $payment_method_storage = \Drupal::entityTypeManager()->getStorage('commerce_payment_method');
      $payment_method = $payment_method_storage->create([
        'type' => $type,
        'payment_gateway' => $payment_gateway->id(),
        'remote_id' => $paymentMethod,
        'uid' => $order->getCustomerId(),
        'billing_profile' => $order->getBillingProfile()
      ]);
      $order->set('payment_method', $payment_method)->save();
      return $payment_method;
    }

    return [
      'type' => 'errors',
      'errors' => [
        [
          'status' => '422',
          'title' => 'Payment Method Error',
          'detail' => 'Billing profile does not exist on cart order!'
        ]
      ]
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getOwnerRemoteCustomerId(OrderInterface $order, string $currency) {
    $customer = $order->getCustomer();

    if (!$customer) {
      return;
    }
    if (!$customer->isAuthenticated()) {
      return;
    }

    // Get the payment config.
    $config_data = $this->getOrderPaymentConfiguration($order, $currency);

    // Retrieve customer remote id.
    $remote_ids = $customer->get('commerce_remote_id');
    $customer_id = $remote_ids->getByProvider($config_data->configName . '|' . $config_data->config['mode']);
    return $customer_id;
  }

  /**
   * {@inheritdoc}
   */
  public function setOwnerRemoteCustomerId(OrderInterface $order, string $currency, string $customer_id) {
    $customer = $order->getCustomer();

    if (!$customer) {
      return;
    }
    if (!$customer->isAuthenticated()) {
      return;
    }

    // Get the payment config.
    $config_data = $this->getOrderPaymentConfiguration($order, $currency);

    // Assign remote customer id to user.
    $remote_ids = $customer->get('commerce_remote_id');
    $remote_ids->setByProvider($config_data->configName . '|' . $config_data->config['mode'], $customer_id);
    $customer->save();
  }

  /**
   * {@inheritdoc}
   */
  public function attachCustomer(OrderInterface $order, $payment_method) {
    // Attach customer to the payment method.
    $currency = strtolower($order->getTotalPrice()->getCurrencyCode());
    $customer_id = $this->getOwnerRemoteCustomerId($order, $currency);
    try {
      $stripe_payment_method = PaymentMethod::retrieve($payment_method);
      if ($customer_id) {
        $stripe_payment_method->attach(['customer' => $customer_id]);
      }
    } catch (ApiErrorException $e) {
      return [
        'type' => 'errors',
        'errors' => [
          [
            'status' => '422',
            'title' => 'Payment Method Error',
            'detail' => 'Failed to update remote payment method.',
          ],
        ],
      ];
    }

    return $stripe_payment_method;
  }

  public function processShipment(OrderInterface $order): object {
    $shipment = $order->get('shipments');
    if ($order->hasField('shipments') && !$order->get('shipments')
        ->isEmpty()) {
      return $shipment->entity;
    }

    $shipment = Shipment::create([
      'type' => 'default',
      'order_id' => $order->id(),
      'title' => 'Shipment',
      'state' => 'ready',
    ]);

    return $shipment;
  }

  public function checkShippingAvailabilityByCountry($order, array $shippingAddress) {
    $blockedCountries = $this->getBlockedShippingCountries($order->getStore()->id());
    $cond1 = (isset($blockedCountries[0]) && $this->getBlockedShippingCountries($order->getStore()->id())[0] == 'none');
    $cond2 = (isset($shippingAddress['country'])
      &&
      !in_array(
        $shippingAddress['country'],
        $this->getBlockedShippingCountries($order->getStore()
          ->id())
      ));

    if ($cond1 || $cond2) {
      return true;
    }

    return false;
  }

  /**
   * get blocked shippied countries
   *
   * @param int $storeId
   *
   * @return array $blockedCountries
   */
  public function getBlockedShippingCountries($storeId): array {
    $blockedCountries = [];
    if (count($this->getAvailableShippingCountries($storeId)) == 1 && $this->getAvailableShippingCountries($storeId)[0] == 'all') {
      $blockedCountries[] = 'none';
    } else {
      $blockedCountries = array_diff(array_keys(CountryManager::getStandardList()), $this->getAvailableShippingCountries($storeId));
    }
    return $blockedCountries;
  }

  /**
   * get all available shipping countries
   *
   * @param int $storeId
   *
   * @return array $availableCountries
   */
  public function getAvailableShippingCountries($storeId): array {
    $store = Store::load($storeId);
    $availableCountries = [];
    if (
      $store->get('shipping_countries') && $store->get('shipping_countries')
        ->isEmpty()
    ) {
      $availableCountries[] = 'all';
    } elseif ($store->get('shipping_countries')) {
      foreach ($store->get('shipping_countries') as $country_item) {
        $availableCountries[] = $country_item->value;
      }
    }
    return $availableCountries;
  }

  /**
   * Converts the given amount to its minor units.
   *
   * For example, 9.99 USD becomes 999.
   *
   * @param \Drupal\commerce_price\Price $amount
   *   The amount.
   *
   * @return int
   *   The amount in minor units, as an integer.
   */
  public function toMinorUnits(Price $amount) {
    $currency_storage = Drupal::entityTypeManager()
      ->getStorage('commerce_currency');
    $currency = $currency_storage->load($amount->getCurrencyCode());
    $fraction_digits = $currency->getFractionDigits();
    $number = $amount->getNumber();
    if ($fraction_digits > 0) {
      $number = Calculator::multiply($number, pow(10, $fraction_digits));
    }

    return intval(round($number, 0));
  }

  public function processCustomerName(string $fullName): array {
    //remove multiple empty spaces to ensure that the last name is saved
    $fullName = preg_replace('/\s+/', ' ', $fullName);
    $names = explode(' ', (string) trim($fullName));
    $given_name = array_shift($names);
    $family_name = implode(' ', $names);
    return [
      'given_name' => $given_name,
      'family_name' => $family_name
    ];
  }
  public function assignOrCreateProfile($order, $paymentMethodData, $type = 'billing') {
    $paymentMethodData = json_decode(json_encode((object) $paymentMethodData), FALSE);
    $paymentMethodData = $paymentMethodData->data;
    // We only get the full name from Apple/Google pay, so we need to "guess" the given
    // name and the family name.

    $currency = strtolower($order->getTotalPrice()->getCurrencyCode());
    $orderPaymentConfiguration = $this->getOrderPaymentConfiguration($order, $currency);

    // Init Stripe and config
    $config = $orderPaymentConfiguration->config;

    $billingDetails = $paymentMethodData->result->paymentMethod->billing_details;
    $shippingDetails = $paymentMethodData->result->shippingAddress; //addressline is a sub-object

    $billingCustomerName = $shippingCustomerName = [];
    $searchData = $createData = [];
    if (
      isset($billingDetails->name) &&
      strlen(trim($billingDetails->name) >= 1)
    ) {
      $billingCustomerName = $this->processCustomerName($billingDetails->name);
    }

    if (
      isset($shippingDetails->recipient) &&
      strlen(trim($shippingDetails->recipient) >= 1)
    ) {
      $shippingCustomerName = $this->processCustomerName($shippingDetails->recipient);
    }

    $givenName = $shippingCustomerName['given_name'];
    $familyName = $shippingCustomerName['family_name'];
    $createData['type'] = $searchData['type'] = 'customer_shipping';
    $createAddressData = [];
    $phone = $shippingDetails->phone;
    if (empty(trim($phone))) {
      $phone = $billingDetails->phone;
    }
    switch ($type) {
      case 'billing';
        $createData['type'] = $searchData['type'] = 'customer';
        //first We check are the legit billing details from the payment method
        if (
          isset($billingCustomerName['given_name']) && !empty($billingCustomerName['given_name']) &&
          isset($billingCustomerName['family_name']) && !empty($billingCustomerName['family_name'])
        ) {
          $givenName = $billingCustomerName['given_name'];
          $familyName = $billingCustomerName['family_name'];
        }
        $createAddressData['address']['country_code'] = $searchData['address.country_code'] = $billingDetails->address->country;
        $createAddressData['address']['locality'] = $searchData['address.locality'] = $billingDetails->address->city;
        $createAddressData['address']['postal_code'] = $searchData['address.postal_code'] = $billingDetails->address->postal_code;
        $createAddressData['address']['address_line1'] = $searchData['address.address_line1'] = $billingDetails->address->line1;
        $createAddressData['address']['address_line2']  = $billingDetails->address->line1;
        if (isset($billingDetails->address->region)) {
          $createAddressData['address']['administrative_area'] = $searchData['address.administrative_area'] = $billingDetails->address->region;
        }
        if (!isset($billingDetails->address->region) && isset($billingDetails->address->state)) {
          $createAddressData['address']['administrative_area'] = $searchData['address.administrative_area'] = $billingDetails->address->state;
        }

        break;
      default: //shipping
        //shipping is already set before switch case so We are double checking if the shipping details are legit
        //if not We use the billing details name
        if (count(array_filter(array_map('trim', $shippingCustomerName))) !== count($shippingCustomerName)) {
          $givenName = $billingCustomerName['given_name'];
          $familyName = $billingCustomerName['family_name'];
        }
        $createAddressData['address']['country_code'] = $searchData['address.country_code'] = $shippingDetails->country;
        $createAddressData['address']['locality'] = $searchData['address.locality'] = $shippingDetails->city;
        $createAddressData['address']['postal_code'] = $searchData['address.postal_code'] = $shippingDetails->postalCode;
        $createAddressData['address']['address_line1'] = $searchData['address.address_line1'] = $shippingDetails->addressLine[0];
        if (isset($shippingDetails->addressLine[1])) {
          $createAddressData['address']['address_line2']  = $shippingDetails->addressLine[1];
        }
        if (isset($shippingDetails->region)) {
          $createAddressData['address']['administrative_area'] = $searchData['address.administrative_area'] = $shippingDetails->region;
        }
    }

    $searchData['address.given_name'] = $givenName;
    $searchData['address.family_name'] = $familyName;

    $profile = false;
    //search the database only if the user is not anonymous
    if ($order->getCustomerId() != 0) {
      //load all profiles to see if shipping address already exists to prevent database bloat
      $searchData['uid'] = $createData['uid'] = $order->getCustomerId();
      //unset empty values because of the loadByProperties
      array_walk_recursive($searchData, function (&$value, $key) {
        if ($value === '') {
          unset($value);  // Unset the value from the array
        }
      });

      $profiles = \Drupal::entityTypeManager()
        ->getStorage('profile')->loadByProperties($searchData);
      $profile = reset($profiles);

      //check if the address exists
    }
    // Merge all fields into a single array.
    $fieldAddressCheck = $createAddressData['address'];
    unset($fieldAddressCheck['administrative_area'], $fieldAddressCheck['address_line2']);

    $allFields = array_merge([$givenName, $familyName], $fieldAddressCheck);
    $profileDataCheck = $this->profileDataCheck($allFields);
    try {
      if (is_bool($profile) && $profileDataCheck) {
        $profile = Profile::create($createData);

        $profile->address->given_name = $givenName;
        $profile->address->family_name = $familyName;
        $profile->address->country_code = $createAddressData['address']['country_code'];
        $profile->address->locality = $createAddressData['address']['locality'];
        $profile->address->administrative_area = $createAddressData['address']['address_line1'];
        $profile->address->postal_code = $createAddressData['address']['postal_code'];
        $profile->address->address_line1 = $createAddressData['address']['address_line1'];
        if (isset($createAddressData['address']['address_line2'])) {
          $profile->address->address_line2 = $createAddressData['address']['address_line2'];
        }
        if (isset($createAddressData['address']['administrative_area'])) {
          $profile->address->administrative_area = $createAddressData['address']['administrative_area'];
        }

        $profile->save();
      }

      if ($profile->hasField('field_phone') && $config['request_phone']) {
        $profile->set('field_phone', $phone);
      }
      $profile->save();
    } catch (\Exception $e) {
      \Drupal::logger('commerce_stripe_payment_request_button')->errror($e->getMessage());
    }

    return $profile;
  }
  public function assignProfileToOrder($order, $profile, string $type = 'billing', $params = []) {
    switch ($type) {
      case 'billing':
        $order->setBillingProfile($profile);
        break;
      default:
        $shipment = null;
        if (!$this->hasShipment($order)) {
          $shipment = $this->createShipment($order, $params);
          $order->set('shipments', $shipment);
          $order->save();
        }
        $shipment = $this->getOrderShipment($order);

        $shipping_method = $shipment->getShippingMethod();
        $shipping_method_plugin = $shipping_method->getPlugin();
        $shipping_rates = $shipping_method_plugin->calculateRates($shipment);
        $shipping_rate = reset($shipping_rates);
        $shipping_service = $shipping_rate->getService()->getId();
        $shipping_method_amount = $shipping_rate->getAmount();

        $shipment->setShippingProfile($profile);
        $shipment->setAmount($shipping_method_amount);
        $shipment->setShippingService($shipping_service);
        $shipment->save();

        $order->setRefreshState(OrderInterface::REFRESH_ON_SAVE);
        break;
    }
    $order->save();
  }

  public function hasShipment($order): bool {
    if ($order->hasField('shipments') && !$order->get('shipments')->isEmpty() && !is_null($order->get('shipments')->entity)) {
      return true;
    }
    return false;
  }

  public function getShipmentByShippingMethod($order, $shippingMethodId) {
    // Get the shipments associated with the order.
    $shipments = Drupal::entityTypeManager()
      ->getStorage('commerce_shipment')
      ->loadByProperties([
        'order_id' => $order->id(),
        'shipping_method' => $shippingMethodId,
      ]);

    // If shipments are found, process them further.
    if (count($shipments)) {
      $shipment = reset($shipments);
      return $shipment;
    }
    return null;
  }

  public function getOrderShipment($order) {
    return $order->get('shipments')->entity;
  }

  public function createShipment($order, $data) {
    $shipment = null;
    $shipment = Shipment::create([
      'type' => 'default',
      'order_id' => $order->id(),
      'title' => 'Shipment',
      'state' => 'ready',
    ]);

    // Loop through order items and add them to shipment
    foreach ($order->getItems() as $order_item) {
      $quantity = $order_item->getQuantity();
      $purchased_entity = $order_item->getPurchasedEntity();


      $weight = new Weight(1, WeightUnit::GRAM);
      if (!$purchased_entity->get('weight')->isEmpty()) {
        $weight_item = $purchased_entity->get('weight')->first();
        $weight = $weight_item->toMeasurement();
      }

      $shipment_item = new ShipmentItem([
        'order_item_id' => $order_item->id(),
        'title' => $purchased_entity->label(),
        'quantity' => $quantity,
        'weight' => $weight->multiply($quantity),
        'declared_value' => $order_item->getTotalPrice(),
      ]);
      $shipment->addItem($shipment_item);
    }

    $shipping_method = null;

    if (!is_null($order->getData('payment_request_button_shipping'))) {
      $shipping_method_storage = \Drupal::entityTypeManager()->getStorage('commerce_shipping_method');
      $shipping_method = $shipping_method_storage->load($order->getData('payment_request_button_shipping'));
    } elseif (isset($data['data']['result']['shippingOption'])) {
      $shipping_method_storage = \Drupal::entityTypeManager()->getStorage('commerce_shipping_method');
      $shipping_method = $shipping_method_storage->load($data['data']['result']['shippingOption']['id']);
    }
    $shipment->setTitle($shipping_method->getName());
    $shipment->setShippingMethodId($shipping_method->id());
    $shipment->setShippingMethod($shipping_method);
    $shipment->save();

    // Add proper shipping method

    $shipping_method_plugin = $shipping_method->getPlugin();
    $shipping_rates = $shipping_method_plugin->calculateRates($shipment);
    $shipping_rate = reset($shipping_rates);
    $shipping_method_amount = $shipping_rate->getAmount();

    //if the order currency is different from the default currency, check the if there is an correct shipping field set
    if (
      $shipping_method_amount->getCurrencyCode() != $order->getTotalPrice()
        ->getCurrencyCode()
      && isset($shipping_method_plugin->getConfiguration()['fields'][$order->getTotalPrice()
          ->getCurrencyCode()])
    ) {
      $newAmount = $shipping_method_plugin->getConfiguration()['fields'][$order->getTotalPrice()
        ->getCurrencyCode()];
      if (is_numeric($newAmount['number']) && $newAmount['number'] > 0) {
        $shipping_method_amount = new Price($newAmount['number'], $newAmount['currency_code']);
      }
    }
    //amount is not zero and there are not currency fields appended We use the commerce_currency_resolver to convert the amount
    elseif (
      $shipping_method_amount->getNumber() > 0 &&
      $shipping_method_amount->getCurrencyCode() != $order->getTotalPrice()
        ->getCurrencyCode()
      && isset($shipping_method_plugin->getConfiguration()['fields'][$order->getTotalPrice()
          ->getCurrencyCode()])
    ) {
      $shipping_method_amount = Drupal::service('commerce_currency_resolver.calculator')
        ->priceConversion($shipping_method_amount, $order->getTotalPrice()
          ->getCurrencyCode());
    }

    return $shipment;
  }


  public function refreshOrder(OrderInterface $order) {
    // Check if we can refresh order.
    Drupal::service('commerce_order.order_refresh')->refresh($order);
  }

  public function profileDataCheck(array $data): bool {
    // Use array_reduce to check if all values are filled.
    return array_reduce(
      $data,
      function ($carry, $value) {
        // Check if the value is an array, then recursively check its values.
        if (is_array($value)) {
          return $carry && $this->areAllFieldsFilled($value);
        }
        // Otherwise, check for empty or null values.
        return $carry && !empty($value) && $value !== 0 && $value !== '0';
      },
      true
    );
  }

  public function areAllFieldsFilled(array $data): bool {
    // Use array_reduce to check if all values are filled.
    return array_reduce($data, function ($carry, $value) {
      // Check if the value is an array, then recursively check its values.
      if (is_array($value)) {
        return $carry && $this->areAllFieldsFilled($value);
      }
      // Otherwise, check for empty or null values.
      return $carry && !empty($value) && $value !== 0 && $value !== '0';
    }, true);
  }

  public function arrayDataCheck($data): bool {

    foreach ($data as $value) {
      if (empty($value) || is_null($value)) {
        return false;
      }
    }
    return true;
  }

  public function shipmentManager() {
    $serviceName = 'commerce_shipping.shipment_manager';
    if (Drupal::hasService($serviceName)) {
      return Drupal::service($serviceName);
    }
    return null;
  }

  /**
   * Evaluates and/or creates new shipment for order.
   *
   * @param \Drupal\commerce_shipping\Entity\ShipmentInterface $shipment
   * @param null $method
   *
   * @return \Drupal\commerce_shipping\Entity\Shipment|null
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function setShipment(Shipment $shipment, $method = NULL) {
    // Load shipment storage.
    $shipping_method_storage = \Drupal::entityTypeManager()->getStorage('commerce_shipping_method');

    /** @var \Drupal\commerce_shipping\Entity\ShippingMethodInterface[] $shipping_methods */
    $shipping_methods = $shipping_method_storage->loadMultipleForShipment($shipment);

    // ShippingMethodStorage::loadMultipleForShipment() uses uasort() that utilises
    // ShippingMethod::sort() function to sort matching shipping methods by weight (ASC).
    // We will use first matching method by weight (ASC) for shipment.
    foreach ($shipping_methods as $shipping_method) {

      // We need to create new shipment only when it is first order shipment,
      // or if we load selected on the checkout shipping method.
      if ($shipping_method->applies($shipment)) {
        if (empty($shipment->getShippingMethod()) || $shipment->getShippingMethod() && $shipping_method->id() === $shipment->getShippingMethod()->id()) {

          /** @var \Drupal\commerce_shipping\Entity\ShippingMethodInterface $shipping_method */
          $shipping_method = \Drupal::service('entity.repository')->getTranslationFromContext($shipping_method);

          // If shipment method matches existing DO NOT create new shipment just return existing.
          if ($method instanceof ShippingMethodInterface && $method->id() === $shipping_method->id()) {
            $order = $shipment->getOrder();
            $order->setData('payment_request_button_shipping', $method);
            $order->save();
            return $shipment;
          }

          $new_shipment = clone $shipment;

          // Get rate.
          $shipping_method_plugin = $shipping_method->getPlugin();

          try {
            $shipping_rates = $shipping_method_plugin->calculateRates($shipment);
          } catch (\Exception $exception) {
            \Drupal::logger('commerce_stripe_payment_request_button')
              ->error('Exception occurred when calculating rates for @name: @message', [
                '@name' => $shipping_method->getName(),
                '@message' => $exception->getMessage(),
              ]);
            continue;
          }

          // Get correct rates with resolved currency.
          // Allow the rates to be altered via code.
          $event = new ShippingRatesEvent($shipping_rates, $shipping_method, $shipment);
          \Drupal::service('event_dispatcher')->dispatch($event, ShippingEvents::SHIPPING_RATES);
          $shipping_rates = $event->getRates();
          $shipping_rate = end($shipping_rates);

          // Set proper data.
          $new_shipment->setShippingMethodId($shipping_rate->getShippingMethodId());
          $new_shipment->setShippingMethod($shipping_method);
          $new_shipment->setAmount($shipping_rate->getAmount());
          $new_shipment->setOriginalAmount($shipping_rate->getOriginalAmount());
          $new_shipment->setShippingService($shipping_rate->getService()->getId());
          $new_shipment->save();
          $order = $shipment->getOrder();
          $order->setData('payment_request_button_shipping', $method);
          $order->save();
          return $new_shipment;
        }
      }
    }

    // No suitable shipments.
    return NULL;
  }

  public function getShippingOptions($order, $shippingAddress = []): array {
    $shippingOptions = [];
    if ($order->hasItems()) {
      //if country not blocked
      if ($this->checkShippingAvailabilityByCountry($order, $shippingAddress)) {
        $shipment = $this->processShipment($order);
        /*
        upon changing the shipping address, We need to set the profiles
        so We can calculate the shipping rates and/or taxes
        */

        // Find rate, we use the first one, pretending there is only one we want to use, fixed rate for order
        $shipping_method_storage = Drupal::entityTypeManager()
          ->getStorage('commerce_shipping_method');
        $shipping_methods = $shipping_method_storage->loadMultipleForShipment($shipment);

        // Loop through shipping methods and add them to array so it can be sent as Response to Stripe
        foreach ($shipping_methods as $shipping_method) {
          $shipping_method_plugin = $shipping_method->getPlugin();
          $shipping_rates = $shipping_method_plugin->calculateRates($shipment);
          $shipping_rate = reset($shipping_rates);
          $shipping_method_amount = $shipping_rate->getAmount();

          //if the order currency is different from the default currency, check the if there is an correct shipping field set
          if (
            $shipping_method_amount->getCurrencyCode() != $order->getTotalPrice()
              ->getCurrencyCode()
            && isset($shipping_method_plugin->getConfiguration()['fields'][$order->getTotalPrice()
                ->getCurrencyCode()])
          ) {
            $newAmount = $shipping_method_plugin->getConfiguration()['fields'][$order->getTotalPrice()
              ->getCurrencyCode()];
            if (is_numeric($newAmount['number']) && $newAmount['number'] > 0) {
              $shipping_method_amount = new Price($newAmount['number'], $newAmount['currency_code']);
            }
          }
          //amount is not zero and there are not currency fields appended We use the commerce_currency_resolver to convert the amount
          elseif (
            $shipping_method_amount->getNumber() > 0 &&
            $shipping_method_amount->getCurrencyCode() != $order->getTotalPrice()
              ->getCurrencyCode()
            && isset($shipping_method_plugin->getConfiguration()['fields'][$order->getTotalPrice()
                ->getCurrencyCode()])
          ) {
            $shipping_method_amount = Drupal::service('commerce_currency_resolver.calculator')
              ->priceConversion($shipping_method_amount, $order->getTotalPrice()
                ->getCurrencyCode());
          }

          $shippingOptions[] = [
            'id' => $shipping_method->id(),
            'label' => $shipping_method->getName(),
            'detail' => $shipping_rate->getService()->getLabel(),
            'amount' => $this->toMinorUnits($shipping_method_amount),
          ];
        }
      }
    }
    //if there are shipping options then arrange them that the preselected is order shipping method
    if (count($shippingOptions)) {
      if ($shipment->getShippingMethodId() !== null) {
        $shippingMethodId = $shipment->getShippingMethodId();
      } else {
        $shippingMethodId = $order->getData('payment_request_button_shipping');
      }
      $shippingOptions = $this->prioritizeShippingOption($shippingOptions, $shippingMethodId);
    }
    return $shippingOptions;
  }

  public function prioritizeShippingOption(&$shippingOptions, $orderShippingId): array {
    $key = array_search($orderShippingId, array_column($shippingOptions, 'id'));
    if ($key !== false) {
      $selectedOption = $shippingOptions[$key];
      unset($shippingOptions[$key]);
      array_unshift($shippingOptions, $selectedOption);
    }
    return $shippingOptions;
  }
}
