<?php

namespace Drupal\commerce_stripe_payment_request_button\Services;

use Drupal\commerce_order\Entity\OrderInterface;
use Stripe\PaymentIntent;
use Drupal\commerce_payment\Entity\PaymentGatewayInterface;

/**
 * Interface PaymentRequestButtonHelperServiceInterface
 *
 * @package Drupal\commerce_stripe_payment_request_button\Services
 */
interface PaymentRequestButtonHelperServiceInterface {

  /**
   * Check if the payment is Google Pay, Apple pay or Link
   *
   * @param PaymentIntent $paymentIntent
   * @return string
   */
  public function getWalletType(PaymentIntent $paymentIntent);

  /**
   * Get the order configuration by order data or currency
   *
   * @param OrderInterface $order
   * @param string $currency
   * @return mixed
   */
  public function getOrderPaymentConfiguration(OrderInterface $order, string $currency);

  /**
   * Update Drupal payment from the payment intent
   *
   * @param OrderInterface $order
   * @param string $intentId
   * @return PaymentIntent
   */
  public function updateDrupalPaymentByPaymentIntent(OrderInterface $order, string $intentId);

  /**
   * Set the following order data: checkout_flow, checkout_step, email, payment_gateway, stripe_intent, wallet
   *
   * @param OrderInterface $order
   * @param PaymentIntent $paymentIntent
   * @param string $email
   * @param string $configName
   * @param string $shipping
   * @return void
   */
  public function setOrderData(OrderInterface $order, PaymentIntent $paymentIntent, string $email, string $configName, string $shipping = '');

  /**
   * Create a Drupal payment from the data provided
   *
   * @param OrderInterface $order
   * @param PaymentGatewayInterface $paymentGateway
   * @param PaymentIntent $paymentIntent
   * @param string $currentUser
   * @return void
   */
  public function createDrupalPayment(OrderInterface $order, PaymentGatewayInterface $paymentGateway, PaymentIntent $paymentIntent, string $currentUser);

  /**
   * Check if the shipping profile exists or create a new one.
   * Afterwards return the profile for processing
   *
   * @param OrderInterface $order
   * @param array $shipping_info
   * @param bool $requestPhone
   * @param bool $payerPhone
   * @return mixed|null
   */
  public function getOrderShippingProfile(OrderInterface $order, array $shipping_info, bool $requestPhone, bool $payerPhone);

  /**
   * Handle shipping related tasks
   *
   * @param OrderInterface $order
   * @param PaymentIntent $paymentIntent
   * @param bool $requestPhone
   * @return void
   */
  public function handleShipping(OrderInterface $order, PaymentIntent $paymentIntent, bool $requestPhone = false);

  /**
   * Process profile for the order
   *
   * @param OrderInterface $order
   * @param PaymentIntent $paymentIntent
   * @param bool $requestPhone
   * @return void
   */
  public function processProfile(OrderInterface $order, PaymentIntent $paymentIntent, bool $requestPhone = true);

  /**
   * Create paymentMethod for the order.
   *
   * @param OrderInterface $order
   * @param PaymentGatewayInterface $payment_gateway
   * @param string $paymentMethod (Stripe payment intent payment method)
   * @return void
   */
  public function createPaymentMethod(OrderInterface $order, PaymentGatewayInterface $payment_gateway, string $paymentMethod);
}
